@@ -1,10 +1,10 @@
 import time
 
-from tapystry import Effect, Strand, Call, Broadcast, Receive, CallFork, First, Cancel, TapystryError, CallThread
+from tapystry import Effect, Strand, Call, Broadcast, Receive, CallFork, First, Cancel, TapystryError, CallThread, Wrapper
 from tapystry import as_effect
 
 
-@as_effect()
+@as_effect("Sequence")
 def Sequence(effects, name=None):
     """Do each of the effects.
     Effects can be a (nested) list/dict structure
@@ -33,7 +33,7 @@ def Sequence(effects, name=None):
     return results
 
 
-@as_effect()
+@as_effect("Join")
 def Join(strands, name=None):
     """
     Returns the result of a strand (or nested structure of strands)
@@ -54,7 +54,7 @@ def Join(strands, name=None):
         return vals
 
 
-def Fork(effects):
+def Fork(effects, *, run_first=False):
     """Do each of the effects in parallel.
     Returns a strand, whose result reflects the same structure as the effects passed in
     """
@@ -65,7 +65,7 @@ def Fork(effects):
 
     def fork_effects(effects):
         if isinstance(effects, Effect):
-            return CallFork(call_fork, (effects,))
+            return CallFork(call_fork, (effects,), run_first=run_first)
         elif isinstance(effects, list):
             return [fork_effects(v) for v in effects]
         else:
@@ -73,11 +73,11 @@ def Fork(effects):
                 raise TapystryError(f"Input to Fork should be an Effect (or nested list/dict of them): {effects}")
             return {k: fork_effects(v) for k, v in effects.items()}
 
-    return Sequence(fork_effects(effects))
+    return Wrapper(Sequence(fork_effects(effects)), type="Fork")
 
 
-@as_effect()
-def Race(effects, name=None):
+@as_effect("Race")
+def Race(effects, name=None, ensure_cancel=True):
     """Wait for the first of the effects to finish.
     Returns a tuple with the key of the winning item, and its value
     """
@@ -91,14 +91,12 @@ def Race(effects, name=None):
     strands = []
     for key, effect in zip(keys, effects_array):
         strand = yield Fork(effect)
-        if strand.is_done():
-            return key, strand.get_result()
         strands.append(strand)
-    i, result = yield First(strands, name=f"Race({name})")
+    i, result = yield First(strands, name=f"Race({name})", ensure_cancel=ensure_cancel)
     return keys[i], result
 
 
-@as_effect(forked=True)
+@as_effect("Subscribe", forked=True)
 def Subscribe(message_key, fn, predicate=None, leading_only=False, latest_only=False):
     """
     Upon receiving any message, runs the specified function on the sent value.
