@@ -1,3 +1,4 @@
+from functools import partial
 import threading
 import queue
 from concurrent.futures import ThreadPoolExecutor
@@ -13,13 +14,14 @@ class Effect(metaclass=abc.ABCMeta):
     """
     Base class for effects which can be yielded to the tapystry event loop.
     """
-    def __init__(self, type, oncancel=(lambda: None), name=None, caller=None, caller_stack_index=2):
+    def __init__(self, type, oncancel=(lambda: None), name=None, caller=None, caller_stack_index=2, immediate=True):
         self.type = type
         self.cancel = oncancel
         self.name = name
         if caller is None:
             caller = inspect.stack()[caller_stack_index]
         self._caller = caller
+        self.immediate = immediate
 
     def __str__(self):
         if self.name is not None:
@@ -27,6 +29,16 @@ class Effect(metaclass=abc.ABCMeta):
         return f"{self.type}"
 
 
+class Wrapper(Effect):
+    """
+    Wrapper around another effect which modifies the type
+    """
+    def __init__(self, effect, type, **effect_kwargs):
+        self.effect = effect
+        super().__init__(type=type, **effect_kwargs)
+
+
+
 class Broadcast(Effect):
     """
     Effect which broadcasts a message for all strands to hear
@@ -34,10 +46,9 @@ class Broadcast(Effect):
     def __init__(self, key, value=None, name=None, immediate=False, **effect_kwargs):
         self.key = key
         self.value = value
-        self.immediate = immediate
         if name is None:
             name = key
-        super().__init__(type="Broadcast", name=name, **effect_kwargs)
+        super().__init__(type="Broadcast", name=name, immediate=immediate, **effect_kwargs)
 
 
 class Receive(Effect):
@@ -72,10 +83,11 @@ class CallFork(Effect):
     Effect which spins up a new strand by calling generator on the specified arguments
     The tapystry engine immediately returns a Strand object.
     """
-    def __init__(self, gen, args=(), kwargs=None, name=None, **effect_kwargs):
+    def __init__(self, gen, args=(), kwargs=None, name=None, run_first=False, **effect_kwargs):
         self.gen = gen
         self.args = args
         self.kwargs = kwargs
+        self.run_first = run_first
         if name is None:
             name = gen.__name__
         super().__init__(type="CallFork", name=name, **effect_kwargs)
@@ -107,9 +119,12 @@ class First(Effect):
     NOTE: Use of this can be dangerous and can lead to deadlocks, as it cancels losers.
           It is safer to us higher-level APIs such as Race and Join
     """
-    def __init__(self, strands, name=None, cancel_losers=True, **effect_kwargs):
+    def __init__(self, strands, name=None, cancel_losers=True, ensure_cancel=None, **effect_kwargs):
         self.strands = strands
         self.cancel_losers = cancel_losers
+        self.ensure_cancel = cancel_losers if ensure_cancel is None else ensure_cancel
+        if not self.cancel_losers:
+            assert not self.ensure_cancel
         if name is None:
             name = ", ".join([str(x) for x in self.strands])
         self.name = name
@@ -228,7 +243,7 @@ class Strand():
             f"  {self._caller.code_context[0].strip()}",
         ]
 
-    def stack(self):
+    def stack(self, indent=0):
         # if self._parent is None:
         #     return [f"Strand[{self.id.hex}]"]
         # else:
@@ -241,9 +256,9 @@ class Strand():
             return s
         else:
             return "\n".join([
-                self._parent.stack(),
-                f"Yields effect {self._parent_effect}, created at",
-                s
+                self._parent.stack(indent=0),
+                " " * indent + f"Yields effect {self._parent_effect}, created at",
+                " " * indent + s
             ])
 
     def _treelines(self, indent=0):
@@ -269,8 +284,6 @@ class Strand():
         # if self._done:  ??
         if self._effect is not None:
             self._effect.cancel()
-        for child in self._live_children:
-            child.cancel()
         self._canceled = True
 
     def is_canceled(self):
@@ -312,13 +325,10 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False, max_threads=Non
     def queue_effect(effect, strand):
         if not isinstance(effect, Effect):
             raise TapystryError(f"Strand yielded non-effect {type(effect)}")
-        if isinstance(effect, Broadcast):
-            if effect.immediate:
-                q.append(_QueueItem(effect, strand))
-            else:
-                q.appendleft(_QueueItem(effect, strand))
-        else:
+        if effect.immediate:
             q.append(_QueueItem(effect, strand))
+        else:
+            q.appendleft(_QueueItem(effect, strand))
 
     def advance_strand(strand, value=_noval):
         if strand.is_canceled():
@@ -346,39 +356,53 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False, max_threads=Non
             return True
         waiting[key].append(receive)
 
-    def add_racing_strand(racing_strands, race_strand, cancel_losers):
+    def cancel_strand(strand):
+        strand.cancel()
+        waiting.pop("done." + strand.id.hex, None)
+        for child in strand._live_children:
+            cancel_strand(child)
+
+    def add_racing_strand(racing_strands, race_strand, cancel_losers, ensure_cancel):
         assert race_strand not in hanging_strands
         hanging_strands.add(race_strand)
 
         received = False
 
-        def receive_fn(i):
-            def receive(val):
-                nonlocal received
-                assert not (cancel_losers and received)
-                if received:
-                    return
-                for j, strand in enumerate(racing_strands):
-                    if j == i:
-                        assert strand.is_done()
-                    else:
+        def declare_winner(i, val):
+            nonlocal received
+            assert not (ensure_cancel and received)
+            if received:
+                return
+            for j, strand in enumerate(racing_strands):
+                if j == i:
+                    assert strand.is_done()
+                else:
+                    if ensure_cancel:
                         assert not strand.is_done()
-                        if cancel_losers:
-                            strand.cancel()
-                received = True
-                assert race_strand in hanging_strands
-                hanging_strands.remove(race_strand)
-                advance_strand(race_strand, (i, val))
-            return receive
+                    if cancel_losers:
+                        cancel_strand(strand)
+            received = True
+            assert race_strand in hanging_strands
+            hanging_strands.remove(race_strand)
+            advance_strand(race_strand, (i, val))
+
+        winner = None
         for i, strand in enumerate(racing_strands):
             if strand.is_done():
-                raise TapystryError(f"Race between effects that are already completed")
-            waiting["done." + strand.id.hex].append(receive_fn(i))
+                if winner is not None and ensure_cancel:
+                    raise TapystryError(f"Race between effects that are already completed")
+                winner = (i, strand)
+        if winner is not None:
+            (i, strand) = winner
+            declare_winner(i, strand.get_result())
+
+        for i, strand in enumerate(racing_strands):
+            waiting["done." + strand.id.hex].append(partial(declare_winner, i))
 
     def resolve_waiting(wait_key, value):
         fns = waiting[wait_key]
         if debug:
-            print("resolving", wait_key, len(fns))
+            print("resolving", wait_key, len(fns), value)
         # clear first in case it mutates
         waiting[wait_key] = [fn for fn in fns if not fn(value)]
 
@@ -408,70 +432,72 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False, max_threads=Non
         thread_strands[id] = strand
         future.add_done_callback(done_callback)
 
-    def handle_item(item):
-        if item.strand.is_canceled():
+    def handle_item(strand, effect):
+        if strand.is_canceled():
             return
 
-        effect = item.effect
         if isinstance(effect, Intercept):
             if not test_mode:
                 raise TapystryError(f"Cannot intercept outside of test mode!")
-            intercepts.append(item)
-            hanging_strands.add(item.strand)
+            intercepts.append((strand, effect))
+            hanging_strands.add(strand)
             return
 
         if test_mode:
             intercepted = False
-            for intercept_item in intercepts:
-                intercept_effect = intercept_item.effect
+            for (intercept_strand, intercept_effect) in intercepts:
                 if intercept_effect.predicate is None or intercept_effect.predicate(effect):
                     intercepted = True
                     break
             if intercepted:
-                hanging_strands.remove(intercept_item.strand)
-                intercepts.remove(intercept_item)
-                hanging_strands.add(item.strand)
-                advance_strand(intercept_item.strand, (effect, make_injector(item.strand)))
+                hanging_strands.remove(intercept_strand)
+                intercepts.remove((intercept_strand, intercept_effect))
+                hanging_strands.add(strand)
+                advance_strand(intercept_strand, (effect, make_injector(strand)))
                 return
 
         if debug:
-            print(f"Handling {effect} (from {item.strand})")
+            print(f"Handling {effect} (from {strand})")
+            print(strand.stack(indent=2))
 
         if not isinstance(effect, Effect):
             raise TapystryError(f"Strand yielded non-effect {type(effect)}")
 
         if isinstance(effect, Broadcast):
             resolve_waiting("broadcast." + effect.key, effect.value)
-            advance_strand(item.strand)
+            advance_strand(strand)
         elif isinstance(effect, Receive):
-            add_waiting_strand("broadcast." + effect.key, item.strand, effect.predicate)
+            add_waiting_strand("broadcast." + effect.key, strand, effect.predicate)
         elif isinstance(effect, Call):
-            strand = Strand(effect._caller, effect.gen, effect.args, effect.kwargs, parent=item.strand, edge=effect.name or "call")
-            if strand.is_done():
+            call_strand = Strand(effect._caller, effect.gen, effect.args, effect.kwargs, parent=strand, edge=effect.name or "call")
+            if call_strand.is_done():
                 # wasn't even a generator
-                advance_strand(item.strand, strand.get_result())
+                advance_strand(strand, call_strand.get_result())
             else:
-                add_waiting_strand("done." + strand.id.hex, item.strand)
-                advance_strand(strand)
+                add_waiting_strand("done." + call_strand.id.hex, strand)
+                advance_strand(call_strand)
         elif isinstance(effect, CallFork):
-            fork_strand = Strand(effect._caller, effect.gen, effect.args, effect.kwargs, parent=item.strand, edge=effect.name or "fork")
-            advance_strand(item.strand, fork_strand)
+            fork_strand = Strand(effect._caller, effect.gen, effect.args, effect.kwargs, parent=strand, edge=effect.name or "fork")
+            if not effect.run_first:
+                advance_strand(strand, fork_strand)
             if not fork_strand.is_done():
                 # otherwise wasn't even a generator
                 advance_strand(fork_strand)
+            if effect.run_first:
+                advance_strand(strand, fork_strand)
         elif isinstance(effect, CallThread):
-            handle_call_thread(effect, item.strand)
+            handle_call_thread(effect, strand)
         elif isinstance(effect, First):
-            add_racing_strand(effect.strands, item.strand, effect.cancel_losers)
+            add_racing_strand(effect.strands, strand, effect.cancel_losers, effect.ensure_cancel)
         elif isinstance(effect, Cancel):
-            effect.strand.cancel()
-            advance_strand(item.strand)
-        elif isinstance(effect, CallThread):
-            advance_strand(item.strand)
+            cancel_strand(effect.strand)
+            advance_strand(strand)
         elif isinstance(effect, DebugTree):
-            advance_strand(item.strand, initial_strand.tree())
+            advance_strand(strand, initial_strand.tree())
+        elif isinstance(effect, Wrapper):
+            handle_item(strand, effect.effect)
         else:
-            raise TapystryError(f"Unhandled effect type {type(effect)}: {item.strand.stack()}")
+            raise TapystryError(f"Unhandled effect type {type(effect)}: {strand.stack()}")
 
     advance_strand(initial_strand)
     while True:
@@ -492,7 +518,7 @@ def run(gen, args=(), kwargs=None, debug=False, test_mode=False, max_threads=Non
 
         if len(q):
             item = q.pop()
-            handle_item(item)
+            handle_item(item.strand, item.effect)
 
     for strand in hanging_strands:
         if not strand.is_canceled():
